﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;

using Windows.Foundation;

using Microsoft.UI.Xaml;
using Microsoft.UI.Xaml.Controls;

namespace TestProject
{
    /// <summary>
    /// 커스텀 가상화 레이아웃
    /// </summary>
    public class CustomVirtualizingLayout : VirtualizingLayout
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 첫번째 인덱스
        /// </summary>
        private int firstIndex = 0;

        /// <summary>
        /// 마지막 인덱스
        /// </summary>
        private int lastIndex = 0;

        /// <summary>
        /// 마지막 이용 가능한 너비
        /// </summary>
        private double lastAvailableWidth = 0.0;

        /// <summary>
        /// 컬럼 오프셋 리스트
        /// </summary>
        private List<double> columnOffsetList = new List<double>();

        /// <summary>
        /// 캐시 테두리 사각형 리스트
        /// </summary>
        private List<Rect> cachedBoundRectangleList = new List<Rect>();

        /// <summary>
        /// 캐시 테두리 사각형 무효화 여부
        /// </summary>
        private bool cachedBoundRectangleInvalid = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 너비 - Width

        /// <summary>
        /// 너비
        /// </summary>
        public double Width { get; set; } = 150;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 항목 변경시 처리하기 (코어) - OnItemsChangedCore(context, source, e)

        /// <summary>
        /// 항목 변경시 처리하기 (코어)
        /// </summary>
        /// <param name="context">컨텍스트</param>
        /// <param name="source">소스</param>
        /// <param name="e">이벤트 인자</param>
        protected override void OnItemsChangedCore(VirtualizingLayoutContext context, object source, NotifyCollectionChangedEventArgs e)
        {
            this.cachedBoundRectangleList.Clear();

            this.firstIndex = 0;
            this.lastIndex  = 0;

            this.cachedBoundRectangleInvalid = true;

            InvalidateMeasure();
        }

        #endregion
        #region 측정하기 (오버라이드) - MeasureOverride(context, availableSize)

        /// <summary>
        /// 측정하기 (오버라이드)
        /// </summary>
        /// <param name="context">컨텍스트</param>
        /// <param name="availableSize">이용 가능한 크기</param>
        /// <returns>측정 크기</returns>
        protected override Size MeasureOverride(VirtualizingLayoutContext context, Size availableSize)
        {
            Rect viewportRectangle = context.RealizationRect;

            if(availableSize.Width != this.lastAvailableWidth || this.cachedBoundRectangleInvalid)
            {
                UpdateCachedBoundRectangle(availableSize);

                this.lastAvailableWidth = availableSize.Width;
            }

            int columnCount = Math.Max(1, (int)(availableSize.Width / Width));

            if(this.columnOffsetList.Count == 0)
            {
                for(int i = 0; i < columnCount; i++)
                {
                    this.columnOffsetList.Add(0);
                }
            }

            this.firstIndex = GetStartIndex(viewportRectangle);

            int currentIndex = firstIndex;

            double nextOffset = -1.0;

            while(currentIndex < context.ItemCount && nextOffset < viewportRectangle.Bottom)
            {
                UIElement childElement = context.GetOrCreateElementAt(currentIndex);

                childElement.Measure(new Size(Width, availableSize.Height));

                if(currentIndex >= cachedBoundRectangleList.Count)
                {
                    int columnIndex = GetIndexOfLowestColumn(columnOffsetList, out nextOffset);

                    this.cachedBoundRectangleList.Add
                    (
                        new Rect
                        (
                            columnIndex * Width,
                            nextOffset,
                            Width,
                            childElement.DesiredSize.Height
                        )
                    );

                    this.columnOffsetList[columnIndex] += childElement.DesiredSize.Height;
                }
                else
                {
                    if(currentIndex + 1 == this.cachedBoundRectangleList.Count)
                    {
                        GetIndexOfLowestColumn(this.columnOffsetList, out nextOffset);
                    }
                    else
                    {
                        nextOffset = this.cachedBoundRectangleList[currentIndex + 1].Top;
                    }
                }

                this.lastIndex = currentIndex;

                currentIndex++;
            }

            Size extentSize = GetExtentSize(availableSize);

            return extentSize;
        }

        #endregion
        #region 배열하기 (오버라이드) - ArrangeOverride(context, finalSize)

        /// <summary>
        /// 배열하기 (오버라이드)
        /// </summary>
        /// <param name="context">컨텍스트</param>
        /// <param name="finalSize">최종 크기</param>
        /// <returns>배열 크기</returns>
        protected override Size ArrangeOverride(VirtualizingLayoutContext context, Size finalSize)
        {
            if(this.cachedBoundRectangleList.Count > 0)
            {
                for(int index = firstIndex; index <= lastIndex; index++)
                {
                    UIElement childElement = context.GetOrCreateElementAt(index);

                    childElement.Arrange(this.cachedBoundRectangleList[index]);
                }
            }

            return finalSize;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 캐시 테두리 사각형 업데이트하기 - UpdateCachedBoundRectangle(availableSize)

        /// <summary>
        /// 캐시 테두리 사각형 업데이트하기
        /// </summary>
        /// <param name="availableSize">이용 가능한 크기</param>
        private void UpdateCachedBoundRectangle(Size availableSize)
        {
            int columnCount = Math.Max(1, (int)(availableSize.Width / Width));

            this.columnOffsetList.Clear();

            for(int i = 0; i < columnCount; i++)
            {
                this.columnOffsetList.Add(0);
            }

            for(int index = 0; index < this.cachedBoundRectangleList.Count; index++)
            {
                int columnIndex = GetIndexOfLowestColumn(this.columnOffsetList, out var nextOffset);

                double previousHeight = this.cachedBoundRectangleList[index].Height;

                this.cachedBoundRectangleList[index] = new Rect(columnIndex * Width, nextOffset, Width, previousHeight);

                this.columnOffsetList[columnIndex] += previousHeight;
            }

            this.cachedBoundRectangleInvalid = false;
        }

        #endregion
        #region 시작 인덱스 구하기 - GetStartIndex(viewportRectangle)

        /// <summary>
        /// 시작 인덱스 구하기
        /// </summary>
        /// <param name="viewportRectangle">뷰포트 사각형</param>
        /// <returns>시작 인덱스</returns>
        private int GetStartIndex(Rect viewportRectangle)
        {
            int startIndex = 0;

            if(this.cachedBoundRectangleList.Count == 0)
            {
                startIndex = 0;
            }
            else
            {
                for(int i = 0; i < this.cachedBoundRectangleList.Count; i++)
                {
                    Rect currentBoundRectangle = this.cachedBoundRectangleList[i];

                    if(currentBoundRectangle.Y < viewportRectangle.Bottom && currentBoundRectangle.Bottom > viewportRectangle.Top)
                    {
                        startIndex = i;

                        break;
                    }
                }
            }

            return startIndex;
        }

        #endregion
        #region 가장 낮은 컬럼 인덱스 구하기 - GetIndexOfLowestColumn(columnOffsetList, lowestOffset)

        /// <summary>
        /// 가장 낮은 컬럼 인덱스 구하기
        /// </summary>
        /// <param name="columnOffsetList">컬럼 오프셋 리스트</param>
        /// <param name="lowestOffset">가장 낮은 오프셋</param>
        /// <returns>가장 낮은 컬럼 인덱스</returns>
        private int GetIndexOfLowestColumn(List<double> columnOffsetList, out double lowestOffset)
        {
            int lowestIndex = 0;

            lowestOffset = columnOffsetList[lowestIndex];

            for(int index = 0; index < columnOffsetList.Count; index++)
            {
                double currentOffset = columnOffsetList[index];

                if(lowestOffset > currentOffset)
                {
                    lowestOffset = currentOffset;

                    lowestIndex = index;
                }
            }

            return lowestIndex;
        }

        #endregion
        #region 범위 크기 구하기 - GetExtentSize(availableSize)

        /// <summary>
        /// 범위 크기 구하기
        /// </summary>
        /// <param name="availableSize">이용 가능한 크기</param>
        /// <returns>범위 크기</returns>
        private Size GetExtentSize(Size availableSize)
        {
            double largestColumnOffset = this.columnOffsetList[0];

            for(int index = 0; index < this.columnOffsetList.Count; index++)
            {
                double currentOffset = this.columnOffsetList[index];

                if(largestColumnOffset < currentOffset)
                {
                    largestColumnOffset = currentOffset;
                }
            }

            return new Size(availableSize.Width, largestColumnOffset);
        }

        #endregion
    }
}